//###########################################################
// File: fat.c
//
// For FAT12, FAT16 and FAT32
// Only for first Partition
// Only for drives with 512 bytes per sector (the most)
//
// Based on a White Paper from MS
// FAT: General Overview of On-Disk Format
// Version 1.03, December 6, 2000
//
// MBR MasterBootRecord
// PC intern 4
// M.Tischer
// Data Becker
//
// 11.10.2006 Replaced "% BYTE_PER_SEC" with "& (BYTE_PER_SEC-1)".
//            Typecast variables from "unsigned long" to "unsigned int" before:
//            secoffset = (unsigned int)fatoffset & (BYTE_PER_SEC-1);
//            Use "unsigned int" for indexing arrays. Does not really speed up,
//            but reduces code size ;)
//
// 25.09.2006 Initialize all global variables to zero for better compatibility
//            with other compilers.
//
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
//#########################################################################
// Last change: 22.10.2006
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: AVR-GCC 3.4.5
//#########################################################################
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "dos.h"

#if defined (FAT_DEBUG_SHOW_FAT_INFO) || defined (FAT_DEBUG_RW_HITS) || defined (FAT_DEBUG_CLUSTERS)
 #include "serial.h" //for testing only
 #include "printf.h" //for testing only
#endif

unsigned char iob[BYTE_PER_SEC];      //file i/o buffer

unsigned long FirstDataSector=0; 
unsigned long FirstRootSector=0; 

unsigned long FATFirstSector=0;  
unsigned char FATtype=0;         

#ifdef USE_FAT32
 unsigned long FAT32RootCluster=0;

 #ifdef FAT_DEBUG_RW_HITS
  unsigned long FATWRHits=0;
  unsigned long FATRDHits=0;
 #endif

#else
 #ifdef FAT_DEBUG_RW_HITS
  U16 FATWRHits=0;
  U16 FATRDHits=0;
 #endif
#endif

unsigned char iob_status=0;
unsigned long FATCurrentSector=0;

#ifdef DOS_WRITE
  unsigned char FATStatus=0; // only for FAT write buffering
#endif

#ifdef USE_FATBUFFER
 unsigned char fatbuf[BYTE_PER_SEC];   //buffer for FAT sectors

#endif

unsigned char secPerCluster=0;

unsigned long RootDirSectors=0;

//use some global variables for file access
#ifdef USE_FAT32
 unsigned long FirstDirCluster=0;
 unsigned long FileFirstCluster=0;     //first cluster of file
 unsigned long FileCurrentCluster=0;   //actual cluster in use
 unsigned long endofclusterchain=0; //value for END_OF_CLUSTERCHAIN 
 unsigned long maxcluster=0;	 // last usable cluster+1
#else
 U16 FirstDirCluster=0;
 U16 FileFirstCluster=0;     //first cluster of file
 U16 FileCurrentCluster=0;   //actual cluster in use
 U16 endofclusterchain=0; //value for END_OF_CLUSTERCHAIN 
 U16 maxcluster=0;	 // last usable cluster+1
#endif

unsigned long FileCurrentSector=0;    //sector with last data read/written
unsigned long File1stClusterSector=0; //1st sector of cluster in use
unsigned long FileClusterCount=0;     //clusters read (not really)
unsigned long FileDirSector=0;        //sector with dir entry of file
unsigned char FileDirOffset=0;        //offset to file entry in FileDirSector
unsigned long FileSize=0;
unsigned long FilePosition=0;         //actual position when reading file         
unsigned char FileFlag=0;		    //read or write
unsigned char FileAttr=0;		    //file attribute
unsigned char FileName[8];
unsigned char FileExt[3];

unsigned long maxsect=0;           // last sector on drive
unsigned long BytesPerCluster=0;   //bytes per cluster

#ifdef USE_FATBUFFER
//############################################################
void UpdateFATBuffer(unsigned long newsector)
//############################################################
{
 if(newsector!=FATCurrentSector) // do we need to update the FAT buffer ?
  {
#ifdef DOS_WRITE
   if(FATStatus>0)
    {
     WriteFATSector(FATCurrentSector,fatbuf); // write the old FAT buffer
     #ifdef FAT_DEBUG_RW_HITS
      FATWRHits++;
     #endif
     FATStatus=0; // flag FAT buffer is save
    } 
#endif
   ReadFATSector(newsector,fatbuf); //read FAT sector
   #ifdef FAT_DEBUG_RW_HITS
    FATRDHits++;
   #endif

   FATCurrentSector=newsector;
  } 
}
#else //#ifdef USE_FATBUFFER

//############################################################
void UpdateFATBuffer(unsigned long newsector)
//############################################################
{
 if(iob_status!=IOB_FAT) // We have to read a FAT sector first if this is true
 {
   ReadFATSector(newsector,iob); //read FAT sector
   #ifdef FAT_DEBUG_RW_HITS
    FATRDHits++;
   #endif
   FATCurrentSector=newsector;
 }

 if(FATCurrentSector!=newsector) // We have to read a FAT sector if this is true
  {
#ifdef DOS_WRITE
   if(FATStatus>0)
    {
     WriteFATSector(FATCurrentSector,iob); // write the old FAT buffer
     #ifdef FAT_DEBUG_RW_HITS
      FATWRHits++;
     #endif
     FATStatus=0; // flag FAT buffer is save
    }
#endif
   ReadFATSector(newsector,iob); //read FAT sector
   #ifdef FAT_DEBUG_RW_HITS
    FATRDHits++;
   #endif

   FATCurrentSector=newsector;
  }
}
#endif //#ifdef USE_FATBUFFER

//############################################################
// get next clusternumber of fat cluster chain
#ifdef USE_FAT32
 unsigned long GetNextClusterNumber(unsigned long cluster)
#else
 U16 GetNextClusterNumber(U16 cluster)
#endif
//############################################################
{
#ifdef USE_FAT12
 U16 tmp, secoffset;
 unsigned char fatoffset;
#endif

 union Convert *cv;

#ifdef FAT_DEBUG_CLUSTERS
#ifdef USE_FAT32
 printf("GNCN %lu\n",cluster);
#else
 printf("GNCN %u\n",cluster);
#endif
#endif

 if(cluster<maxcluster) //we need to check this ;-)
  {

#ifdef USE_FAT12
   if(FATtype==FAT12)
    {
     // FAT12 has 1.5 Bytes per FAT entry
     // FAT12 can only have 4085 clusters. So cluster * 3 is 16 bit 
     tmp = ((U16)cluster * 3) >>1 ; //multiply by 1.5 (rounds down)
     secoffset = (U16)tmp & (BYTE_PER_SEC-1); //we need this for later
     // FAT12 4085 Cluster * 1.5Bytes = 6127.5 Bytes => max 12 FAT sectors
     // FAT sector offset is 8 Bit 
     fatoffset = (unsigned char)(tmp / BYTE_PER_SEC); //sector offset from FATFirstSector
 
     UpdateFATBuffer(FATFirstSector + fatoffset); //read FAT sector

#ifdef USE_FATBUFFER
     if(secoffset == (BYTE_PER_SEC-1)) //if this is the case, cluster number is
                                   //on a sector boundary. read the next sector too
      {
       tmp=(U16)fatbuf[BYTE_PER_SEC-1]; //keep first part of cluster number
       UpdateFATBuffer(FATFirstSector + fatoffset +1 ); //read next FAT sector
       tmp+=(U16)fatbuf[0] << 8; //second part of cluster number
      }
     else
      {
       cv=(union Convert *)&fatbuf[secoffset];
       tmp=cv->ui;
      } 
#else //#ifdef USE_FATBUFFER

     if(secoffset == (BYTE_PER_SEC-1)) //if this is the case, cluster number is
                                   //on a sector boundary. read the next sector too
      {
       tmp=(U16)iob[BYTE_PER_SEC-1]; //keep first part of cluster number
   
       UpdateFATBuffer(FATFirstSector + fatoffset +1); //read next FAT sector

       tmp+=(U16)iob[0] << 8; //second part of cluster number
      }
     else
      {
       cv=(union Convert *)&iob[secoffset];
       tmp=cv->ui;
      } 
#endif //#ifdef USE_FATBUFFER

     if((unsigned char)cluster & 0x01) tmp>>=4; //shift to right position
     else               tmp&=0xFFF; //delete high nibble

     return (tmp);
    }//if(FATtype==FAT12)
#endif //#ifdef USE_FAT12

#ifdef USE_FAT16
   if(FATtype==FAT16)
    {
     //two bytes per FAT entry
//     UpdateFATBuffer(FATFirstSector + ((unsigned long)cluster * 2) / BYTE_PER_SEC);
//     UpdateFATBuffer(FATFirstSector + ((unsigned int)cluster) / (BYTE_PER_SEC/2));

     UpdateFATBuffer(FATFirstSector + (unsigned char)( (U16)cluster >> 8 ));
#ifdef USE_FATBUFFER
     // Buffer index is max 511. So in any case we loose upper bits. typecast cluster to UINT
//     cv=(union Convert *)&fatbuf[((unsigned long)cluster * 2) % BYTE_PER_SEC];
     cv=(union Convert *)&fatbuf[((U16)cluster << 1) & (BYTE_PER_SEC-1)];
#else //#ifdef USE_FATBUFFER
//     cv=(union Convert *)&iob[((unsigned long)cluster * 2) % BYTE_PER_SEC];
//     cv=(union Convert *)&iob[((unsigned int)cluster << 1) % BYTE_PER_SEC];
     cv=(union Convert *)&iob[((U16)cluster << 1) & (BYTE_PER_SEC-1)];
#endif //#ifdef USE_FATBUFFER

     return(cv->ui);
    }//if(FATtype==FAT16)
#endif //#ifdef USE_FAT16

#ifdef USE_FAT32
   if(FATtype==FAT32)
    {
     //four bytes per FAT entry
//     UpdateFATBuffer(FATFirstSector + (cluster * 4) / BYTE_PER_SEC);
//     UpdateFATBuffer(FATFirstSector + cluster / (BYTE_PER_SEC/4));

     UpdateFATBuffer(FATFirstSector + ( cluster >> 7 ));
#ifdef USE_FATBUFFER
     // Buffer index is max 511. So in any case we loose upper bits. typecast cluster to UINT
     cv=(union Convert *)&fatbuf[((U16)cluster << 2) & (BYTE_PER_SEC-1)];
#else //#ifdef USE_FATBUFFER
     cv=(union Convert *)&iob[((U16)cluster << 2) & (BYTE_PER_SEC-1)];
#endif //#ifdef USE_FATBUFFER

     return( cv->ul & 0x0FFFFFFF );
    }//if(FATtype==FAT32) 
#endif //#ifdef USE_FAT32
  }
  
 return DISK_FULL; //return impossible cluster number
}

//###########################################################
//first datacluster is cluster 2 !
#ifdef USE_FAT32
 unsigned long GetFirstSectorOfCluster(unsigned long n)
#else
 unsigned long GetFirstSectorOfCluster(U16 n)
#endif
//###########################################################
{
 return (((unsigned long)(n - 2) * secPerCluster) + FirstDataSector);
}

#ifdef DOS_WRITE
//###########################################################
#ifdef USE_FAT32
 unsigned long AllocCluster(unsigned long currentcluster)
#else
 U16 AllocCluster(U16 currentcluster)
#endif
//###########################################################
{
#ifdef USE_FAT32
 unsigned long cluster;
#else
 U16 cluster;
#endif

// do this if you want to search from beginning of FAT
// cluster=FindFreeCluster(0); //get next free cluster number
 cluster=FindFreeCluster(currentcluster); // get next free cluster number
 if(cluster!=DISK_FULL && cluster<=maxcluster) // disk full ?
  {
    // insert new cluster number into chain
    // currentcluster=0 means: this is a new cluster chain
    if(currentcluster>0) WriteClusterNumber(currentcluster,cluster);

   // mark end of cluster chain
#ifdef USE_FAT12
   if(FATtype==FAT12) WriteClusterNumber(cluster,0xFFF); 
#endif
#ifdef USE_FAT16
   if(FATtype==FAT16) WriteClusterNumber(cluster,0xFFFF); 
#endif
#ifdef USE_FAT32
   if(FATtype==FAT32) WriteClusterNumber(cluster,0x0FFFFFFF); 
#endif
  }

#ifdef USE_FATBUFFER
#else
   // We should flush the temporary fatbuffer now because next action is
   // filling iob[] with new data !
   if(FATStatus>0)
    {
     WriteFATSector(FATCurrentSector,iob); // write the FAT buffer
     #ifdef FAT_DEBUG_RW_HITS
      FATWRHits++;
     #endif
     FATStatus=0; // flag FAT buffer is save
    } 
#endif

 return cluster;
}
#endif //DOS_WRITE

#ifdef DOS_WRITE
//###########################################################
// go through the FAT to find a free cluster
#ifdef USE_FAT32
 unsigned long FindFreeCluster(unsigned long currentcluster)
#else
 U16 FindFreeCluster(U16 currentcluster)
#endif
//###########################################################
{
#ifdef USE_FAT32
 unsigned long cluster;
#else
 U16 cluster;
#endif

 cluster=currentcluster+1; // its a good idea to look here first
                           // maybe we do not need to search the whole FAT
                           // and can speed up free cluster search
                           // if you do not want this call FindFreeCluster(0)
// search til end of FAT
 while(cluster<maxcluster)
  {
   if(GetNextClusterNumber(cluster)==0) break;
   cluster++;
  }

// if we have not found a free cluster til end of FAT
// lets start a new search at beginning of FAT
 if(cluster>=maxcluster)
  {
   cluster=2; // first possible free cluster
   while(cluster<=currentcluster) // search til we come to where we have started
    {
     if(GetNextClusterNumber(cluster)==0) break;
     cluster++;
    }

   if(cluster>=currentcluster) return DISK_FULL; // no free cluster found
  }
  
 if(cluster>=maxcluster) return DISK_FULL;
    
 return cluster;
}
#endif //DOS_WRITE

#ifdef DOS_WRITE
//############################################################
// write a cluster number into fat cluster chain
#ifdef USE_FAT32
 unsigned char WriteClusterNumber(unsigned long cluster, unsigned long number)
#else
 unsigned char WriteClusterNumber(U16 cluster, U16 number)
#endif
//############################################################
{
#ifdef USE_FAT12
 U16 tmp, secoffset;
 unsigned char fatoffset;
 unsigned char lo,hi;
#endif

 unsigned long sector;
 unsigned char *p;
 
#ifdef FAT_DEBUG_CLUSTERS
#ifdef USE_FAT32
 printf("WCN %lu\n",cluster);
#else
 printf("WCN %u\n",cluster);
#endif
#endif

 if(cluster<maxcluster) //we need to check this ;-)
  {

#ifdef USE_FAT12
   if(FATtype==FAT12)
    {
     //FAT12 has 1.5 Bytes per FAT entry
     // FAT12 can only have 4085 clusters. So cluster * 3 is 16 bit 
     tmp= ((U16)cluster * 3) >>1 ; //multiply by 1.5 (rounds down)
     secoffset = (U16)tmp & (BYTE_PER_SEC-1); //we need this for later
     // FAT12 4085 Cluster * 1.5Bytes = 6127.5 Bytes => max 12 FAT sectors
     // FAT sector offset is 8 Bit 
     fatoffset = (unsigned char)(tmp / BYTE_PER_SEC); //sector offset from FATFirstSector
     sector=FATFirstSector + fatoffset;

     tmp=(U16)number;
     if((unsigned char)cluster & 0x01) tmp<<=4; //shift to right position
     lo=(unsigned char)tmp;
     hi=(unsigned char)(tmp>>8);
     
     UpdateFATBuffer(sector); //read FAT sector

     if(secoffset == (BYTE_PER_SEC-1)) //if this is the case, cluster number is
                                   //on a sector boundary. read the next sector too
      {
#ifdef USE_FATBUFFER
       p=&fatbuf[BYTE_PER_SEC-1]; //keep first part of cluster number
#else //#ifdef USE_FATBUFFER
       p=&iob[BYTE_PER_SEC-1]; //keep first part of cluster number
#endif //#ifdef USE_FATBUFFER

       if((unsigned char)cluster & 0x01)
        {
         *p&=0x0F;
         *p|=lo;
        }
       else *p=lo;
        
       FATStatus=1; // we have made an entry, so write before next FAT sector read
       UpdateFATBuffer(sector+1); //read FAT sector

#ifdef USE_FATBUFFER
       p=&fatbuf[0]; //second part of cluster number
#else //#ifdef USE_FATBUFFER
       p=&iob[0]; //second part of cluster number
#endif //#ifdef USE_FATBUFFER

       if((unsigned char)cluster & 0x01) *p=hi;
       else
        {
         *p&=0xF0;
         *p|=hi;
        }

       FATStatus=1; // we have made an entry, so write before next FAT sector read
      }
     else
      {
#ifdef USE_FATBUFFER
       p=&fatbuf[secoffset];
#else //#ifdef USE_FATBUFFER
       p=&iob[secoffset];
#endif //#ifdef USE_FATBUFFER

       if((unsigned char)cluster & 0x01)
        {
         *p&=0x0F;
         *p++|=lo;
         *p=hi;
        } 
       else
        {
         *p++=lo;
         *p&=0xF0;
         *p|=hi;
        } 

       FATStatus=1; // we have made an entry, so write before next FAT sector read
      } 

    }//if(FATtype==FAT12)
#endif

#ifdef USE_FAT16
   if(FATtype==FAT16)
    {
     //two bytes per FAT entry
//     sector=FATFirstSector + ((unsigned long)cluster * 2) / BYTE_PER_SEC;
//     sector=FATFirstSector + ((unsigned int)cluster) / (BYTE_PER_SEC/2);
//     UpdateFATBuffer(sector); //read FAT sector
     UpdateFATBuffer(FATFirstSector + (unsigned char)( (U16)cluster >> 8 ));

#ifdef USE_FATBUFFER
     // Buffer index is max 511. So in any case we loose upper bits. typecast cluster to UINT
     p=&fatbuf[((U16)cluster << 1) & (BYTE_PER_SEC-1)];
#else //#ifdef USE_FATBUFFER
     p=&iob[((U16)cluster << 1) & (BYTE_PER_SEC-1)];
#endif //#ifdef USE_FATBUFFER

     *p++=(unsigned char)(number);
     *p  =(unsigned char)(number >> 8);

     FATStatus=1; // we have made an entry, so write before next FAT sector read
    }// if(FATtype==FAT16)
#endif //#ifdef USE_FAT16

#ifdef USE_FAT32
   if(FATtype==FAT32)
    {
     //four bytes per FAT entry
//     sector=FATFirstSector + (cluster * 4) / BYTE_PER_SEC;
//     sector=FATFirstSector + cluster / (BYTE_PER_SEC/4);
//     UpdateFATBuffer(sector); //read FAT sector
     UpdateFATBuffer(FATFirstSector + ( cluster >> 7 ));

#ifdef USE_FATBUFFER
     // Buffer index is max 511. So in any case we loose upper bits. typecast cluster to UINT
     p=&fatbuf[((U16)cluster << 2) & (BYTE_PER_SEC-1)];
#else //#ifdef USE_FATBUFFER
     p=&iob[((U16)cluster << 2) & (BYTE_PER_SEC-1)];
#endif //#ifdef USE_FATBUFFER

     number&=0x0FFFFFFF;
     
     *p++=(unsigned char)( number);
     *p++=(unsigned char)(number >> 8);
     *p++=(unsigned char)(number >> 16);
     *p  =(unsigned char)(number >> 24);

     FATStatus=1; // we have made an entry, so write before next FAT sector read

    }// if(FATtype==FAT32) 
#endif //#ifdef USE_FAT32
  } // if(cluster<maxcluster) //we need to check this ;-)

 return 0;
}
#endif //DOS_WRITE

//###########################################################
unsigned char GetDriveInformation(void)
//###########################################################
{
 unsigned char by;
 unsigned long DataSec,TotSec;
 unsigned long bootSecOffset;
 unsigned long FATSz; // FATSize
 unsigned char loop;

#ifdef USE_FAT32
  unsigned long CountofClusters;
#else
  U16 CountofClusters;
#endif

 U16  RootEntrys;

 struct MBR *mbr;
 struct BootSec *boot;
  
 by=IdentifyMedia(); //LaufwerksInformationen holen
 if(by==0)
  {
   FATtype=0; //Unknown FAT type
   bootSecOffset=0; //erstmal

   by=ReadSector(0,iob); //Lese den MBR. Erster Sektor auf der Platte
                       //enthlt max. 4 Partitionstabellen mit jeweils 16Bytes
                       //Die erste fngt bei 0x01BE an, und nur die nehme ich !

   //Erstmal checken ob wir nicht schon einen Bootsektor gelesen haben.
   boot=(struct BootSec *)iob;

   loop=0;
   do
    {
     // Jetzt checke ich doch den FAT-String im Bootsektor um den Typ der FAT
     // zu bestimmen. Einen besseren Weg sehe ich im Moment nicht.
     if(   boot->eb.rm.BS_FilSysType[0]=='F'
//        && boot->eb.rm.BS_FilSysType[1]=='A'
//        && boot->eb.rm.BS_FilSysType[2]=='T'
        && boot->eb.rm.BS_FilSysType[3]=='1'  )
      {
       //Wenn ich hier ankomme habe ich entweder FAT12 oder FAT16
#ifdef USE_FAT12
       if(boot->eb.rm.BS_FilSysType[4]=='2') FATtype=FAT12;
#endif
#ifdef USE_FAT16
       if(boot->eb.rm.BS_FilSysType[4]=='6') FATtype=FAT16;
#endif
      }
     else
      {
#ifdef USE_FAT32
       if(   boot->eb.rm32.BS_FilSysType[0]=='F'
//          && boot->eb.rm32.BS_FilSysType[1]=='A'
//          && boot->eb.rm32.BS_FilSysType[2]=='T'
          && boot->eb.rm32.BS_FilSysType[3]=='3'
          && boot->eb.rm32.BS_FilSysType[4]=='2'
         )
        {
         FATtype=FAT32;
        }
       else //war kein Bootsektor, also feststellen wo der liegt
#endif
        {
         mbr=(struct MBR *)iob; //Pointer auf die Partitionstabelle
         bootSecOffset=mbr->part1.bootoffset; //Nur den brauche ich

         by=ReadSector(bootSecOffset,iob);      //read bootsector
         boot=(struct BootSec *)iob;
        } 
      }

     loop++;
    }while(loop<2 && FATtype==0); //Bis zu zwei Versuche den Bootsektor zu lesen

   if(FATtype==0) return F_ERROR; // FAT-Typ nicht erkannt

   secPerCluster=boot->BPB_SecPerClus; //Sectors per Cluster
   RootEntrys=boot->BPB_RootEntCnt;    //32 Byte Root Directory Entrys

//Number of sectors for FAT
   if(boot->BPB_FATSz16 != 0) FATSz = boot->BPB_FATSz16;
   else FATSz = boot->eb.rm32.BPB_FATSz32; //Fr FAT32

   RootDirSectors = ((RootEntrys * 32) + (BYTE_PER_SEC - 1)) / BYTE_PER_SEC;
   
   FATFirstSector= bootSecOffset + boot->BPB_RsvdSecCnt;
   FirstRootSector = FATFirstSector + (boot->BPB_NumFATs * FATSz);
   FirstDataSector = FirstRootSector + RootDirSectors;

   if(boot->BPB_TotSec16 != 0) TotSec = boot->BPB_TotSec16;
   else TotSec = boot->BPB_TotSec32;

//Number of data sectors
   DataSec = TotSec - (boot->BPB_RsvdSecCnt + (boot->BPB_NumFATs * FATSz) + RootDirSectors);

//Number of valid clusters
   CountofClusters = DataSec / secPerCluster;
   maxcluster=CountofClusters+2;

//Note also that the CountofClusters value is exactly that: the count of data clusters
//starting at cluster 2. The maximum valid cluster number for the volume is
//CountofClusters + 1, and the "count of clusters including the two reserved clusters"
// is CountofClusters + 2.

   FirstDirCluster=0; // for FAT12 and FAT16

#ifdef USE_FAT12
   if(FATtype==FAT12)
    {
     endofclusterchain=EOC12;
    } 
#endif
#ifdef USE_FAT16
   if(FATtype==FAT16)
    {
     endofclusterchain=EOC16;
    } 
#endif
#ifdef USE_FAT32
   if(FATtype==FAT32)
    {
     endofclusterchain=EOC32;
     FAT32RootCluster=boot->eb.rm32.BPB_RootClus;
     FirstDirCluster=FAT32RootCluster;
     FirstRootSector=GetFirstSectorOfCluster(FAT32RootCluster);
    }
#endif

  }
 else
  {
   return F_ERROR; // CF gives no answer
  } 

 FileFirstCluster=0;
 FileSize=0;
 FileFlag=0;
 BytesPerCluster=BYTE_PER_SEC * secPerCluster; //bytes per cluster

 FATCurrentSector=FATFirstSector;

// for debugging only
#ifdef FAT_DEBUG_SHOW_FAT_INFO
 if(FATtype==FAT12) puts("FAT12\n");
 if(FATtype==FAT16) puts("FAT16\n");
 if(FATtype==FAT32) puts("FAT32\n");

 printf("bootSecOffset %lu\n",bootSecOffset);
 printf("Reserved Sectors %u\n",boot->BPB_RsvdSecCnt);
 printf("FAT Sectors %lu\n",FATSz);
 printf("Num. of FAT's %u\n",(U16)boot->BPB_NumFATs);
 printf("secPerCluster %u\n",(U16)secPerCluster);
 printf("BytesPerCluster %lu\n",BytesPerCluster);
 printf("FATFirstSector %lu\n",FATFirstSector);
 printf("FirstRootSector %lu\n",FirstRootSector);
 printf("RootDirSectors %lu\n",RootDirSectors);
 printf("FirstDataSector %lu\n",FirstDataSector);
 printf("maxsect %lu\n",maxsect);
#ifdef USE_FAT32
 printf("FirstDirCluster %lu\n",FirstDirCluster);
 printf("maxcluster %lu\n",maxcluster);
#else
 printf("FirstDirCluster %u\n",FirstDirCluster);
 printf("maxcluster %u\n",maxcluster);
#endif

#endif //#ifdef FAT_DEBUG_SHOW_FAT_INFO

 #ifdef DOS_WRITE
  FATStatus=0; // nothing to write til here
 #endif

#ifdef USE_FATBUFFER
 ReadFATSector(FATCurrentSector,fatbuf); //read first FAT sector
#else
 ReadFATSector(FATCurrentSector,iob); //read first FAT sector
#endif


 return F_OK;
}

#ifndef USE_FATBUFFER
#ifdef DOS_WRITE
//#########################################################################
unsigned char WriteFATSector(unsigned long sector, unsigned char *buf)
//#########################################################################
{
 iob_status=IOB_FAT; // we have to track this
 return WriteSector(sector,buf); //write sector
}
#endif

//#########################################################################
unsigned char ReadFATSector(unsigned long sector, unsigned char *buf)
//#########################################################################
{
 iob_status=IOB_FAT; // we have to track this
 return ReadSector(sector,buf); //read sector
}

#ifdef DOS_WRITE // Better keep this as a function !
//#########################################################################
unsigned char WriteDirSector(unsigned long sector, unsigned char *buf)
//#########################################################################
{
 iob_status=IOB_DIR; // we have to track this
 return WriteSector(sector,buf); //write sector
}
#endif

//#########################################################################
unsigned char ReadDirSector(unsigned long sector, unsigned char *buf)
//#########################################################################
{
 iob_status=IOB_DIR; // we have to track this
 return ReadSector(sector,buf); //read sector
}

#ifdef DOS_WRITE // This has to be a function for Fwrite(), Fclose() !
//#########################################################################
unsigned char WriteFileSector(unsigned long sector, unsigned char *buf)
//#########################################################################
{
 iob_status=IOB_DATA; // we have to track this
 return WriteSector(sector,buf); //write sector
}
#endif

//#########################################################################
unsigned char ReadFileSector(unsigned long sector, unsigned char *buf)
//#########################################################################
{
 iob_status=IOB_DATA; // we have to track this
 return ReadSector(sector,buf); //read sector
}

#endif //#ifndef USE_FATBUFFER
