//#########################################################################
// File: fat.h
//
// Benutzt nur die erste Partition
// Nur fr Laufwerke mit 512 Bytes pro Sektor
//
// Nach einem White Paper von MS
// FAT: General Overview of On-Disk Format
// Version 1.03, December 6, 2000
//
// 28.12.2006 Removed "unsigned int" ! On ARM this is 4 bytes. Added typedefs for U16.
//            Added defined(__arm__) || defined(__thumb__) for ARM.
//            Added __attribute__((packed)) in FAT struct's for ARM.
//            Added defined(__AVR__) || defined(ATMEGA)    for AVR.
//
// 24.09.2006 Typecast most constants for better compatibility with other compilers
//            like MCC18 for PIC18Fxxxx
//
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
//#########################################################################
// Last change: 28.12.2006
//#########################################################################
// Compiler: WinAVR 3.4.5
// Compiler: WinARM
//#########################################################################

#ifndef __FAT_H
#define __FAT_H

#if defined(__arm__) || defined(__thumb__)

 #include "typedefs.h" // ARM specific types

#elif defined(__AVR__) || defined(ATMEGA)

 typedef unsigned char U8;  // one byte
 typedef unsigned int U16;   // two bytes
 typedef unsigned long U32;  // four bytes

#endif

// These defines are for FAT debugging only. You don't need them.
// Activating all three options takes about 1.5kB of flash.
// So be careful on devices with small flash memory !
//#define FAT_DEBUG_SHOW_FAT_INFO //activate FAT information output via printf() or puts() in GetDriveInformation()
//#define FAT_DEBUG_CLUSTERS // show cluster numbers read/write access via printf()
//#define FAT_DEBUG_RW_HITS // show FAT sectors read/write access summary when calling Fclose(), via printf()

//file operations
#define END_DIR		0
#define NO_MATCH	1
#define MATCH_NAME	2
#define MATCH_EXT	3
#define FULL_MATCH	MATCH_NAME + MATCH_EXT

#define IOB_DATA	(U8)0x01  // iob[] holds data of a file sector
#define IOB_FAT		(U8)0x02  // iob[] holds data of a FAT sector
#define IOB_DIR		(U8)0x03  // iob[] holds data of a directory sector

#define PART1_TABLE_OFFSET (U16) 0x01BE //offset to first partitiontable in sector 0

//Using structures needs less memory than indexing in arrays like inbuff[]

//partitiontable structure
//most of it is not used in this program
//bootsector offset is the only thing we need
//because C/H/S values are not used. LBA !
struct PartInfo {
                 U8 status;      //Partition status, 0x80 = Active, 0x00 = inactive
                 U8 firsthead;   //First head used by partition
                 U16           firstseccyl; //First sector and cylinder used by partition
                 U8 type;        //Partition type
                 U8 lasthead;    //Last head used by partition
                 U16           lastseccyl;  //Last sector and cylinder used by partition
                 U32 bootoffset;  //Location of boot sector. !!!!!!!!!!!
                 U32 secofpart;   //Number of sectors for partition
} __attribute__((packed));

//first sector of disc is the master boot record
//it contains four partitiontables
//only the first partition is used in this program
struct MBR {
            U8 dummy[PART1_TABLE_OFFSET]; //we don't need all these bytes
            struct PartInfo part1;
            struct PartInfo part2;
            struct PartInfo part3;
            struct PartInfo part4;
//all bytes below are not necessary
} __attribute__((packed));

//part of FAT12/16 bootsector different to FAT32
struct RemBoot //FAT12/16 defs beginning at offset 36
 {
	U8  BS_DrvNum;
	U8  BS_Reserved1;
	U8  BS_BootSig;
	U8  BS_VolID[4];
	char           BS_VolLab[11];
	char           BS_FilSysType[8];
	U8  remaining_part[450];
 } __attribute__((packed));

//part of FAT32 bootsector different to FAT12/16
struct RemBoot32 //FAT32 defs beginning at offset 36
  {
	U32            BPB_FATSz32; //4 bytes
	U16            BPB_ExtFlags; //2 bytes
	U16            BPB_FSVer; //2 bytes
	U32            BPB_RootClus; //4 bytes
	U16            BPB_FSInfo; //2 bytes
	U16            BPB_BkBootSec; //2 bytes
	U8             BPB_Reserved[12];
	U8             BS_DrvNum;
	U8             BS_Reserved1;
	U8             BS_BootSig;
	U32            BS_VolID; //4 bytes
	char           BS_VolLab[11];
	char           BS_FilSysType[8];
	U8             remaining_part[422];
} __attribute__((packed)); 

union endboot 
{
       struct RemBoot   rm;
       struct RemBoot32 rm32;
} __attribute__((packed));

struct BootSec 
{
	unsigned char  BS_jmpBoot[3];
	char           BS_OEMName[8];
	U16            BPB_BytesPerSec; //2 bytes
	unsigned char  BPB_SecPerClus;
	U16            BPB_RsvdSecCnt; //2 bytes
	unsigned char  BPB_NumFATs;
	U16            BPB_RootEntCnt; //2 bytes
	U16            BPB_TotSec16; //2 bytes
	unsigned char  BPB_Media;
	U16            BPB_FATSz16; //2 bytes
	U16            BPB_SecPerTrk; //2 bytes
	U16            BPB_NumHeads; //2 bytes
	unsigned long  BPB_HiddSec; //4 bytes
	unsigned long  BPB_TotSec32; //4 bytes
        union endboot  eb; //remaining part of bootsector
} __attribute__((packed));


#define BYTE_PER_SEC	(U16) 512

#define FAT12	(unsigned char) 12
#define FAT16	(unsigned char) 16
#define FAT32	(unsigned char) 32

//defines for special cluster values
//free cluster has value 0
//for fat32 don't use upper four bits ! ignore them
//cluster value of 0x10000000 is a FREE cluster in FAT32
 
//values for end of cluster chain
//ranges for example for FAT12 from 0xFF8 to 0xFFF
#define EOC12	(U16)           0xFF8
#define EOC16	(U16)           0xFFF8
#define EOC32	(unsigned long) 0x0FFFFFF8

//values for bad marked clusters
#define BADC12	(U16)           0xFF7
#define BADC16	(U16)           0xFFF7
#define BADC32	(unsigned long) 0x0FFFFFF7

//values for reserved clusters
//ranges for example for FAT12 from 0xFF0 to 0xFF6
#define RESC12	(U16)           0xFF0
#define RESC16	(U16)           0xFFF0
#define RESC32	(unsigned long) 0x0FFFFFF0

#ifdef USE_FAT32
 #define DISK_FULL (unsigned long) 0xFFFFFFFF
#else                              
 #define DISK_FULL (U16)           0xFFFF
#endif

//File/Dir Attributes
#define ATTR_FILE	(unsigned char)0x00 //not defined by MS ! I did it 
#define ATTR_READ_ONLY	(unsigned char)0x01
#define ATTR_HIDDEN	(unsigned char)0x02
#define ATTR_SYSTEM	(unsigned char)0x04
#define ATTR_VOLUME_ID	(unsigned char)0x08
#define ATTR_DIRECTORY	(unsigned char)0x10
#define ATTR_ARCHIVE	(unsigned char)0x20
#define ATTR_LONG_NAME  (unsigned char)0x0F
#define ATTR_NO_ATTR  	(unsigned char)0xFF //not defined by MS ! I did it 

//Char codes not allowed in a filename
//NOT checked yet
//0x22, 0x2A, 0x2B, 0x2C, 0x2E, 0x2F, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F, 0x5B, 0x5C, 0x5D, and 0x7C. 

struct DirEntry {
                 unsigned char DIR_Name[8];      //8 chars filename
                 unsigned char DIR_Ext[3];       //3 chars extension
                 unsigned char DIR_Attr;         //file attributes RSHA
                 unsigned char DIR_NTres;        //set to zero
                 unsigned char DIR_CrtTimeTenth; //creation time part in milliseconds
                 U16           DIR_CrtTime;      //creation time
                 U16           DIR_CrtDate;      //creation date
                 U16           DIR_LastAccDate;  //last access date (no time for this !)
                 U16           DIR_FstClusHI;  //first cluster high word                 
                 U16           DIR_WrtTime;      //last write time
                 U16           DIR_WrtDate;      //last write date
                 U16           DIR_FstClusLO;  //first cluster low word                 
                 unsigned long DIR_FileSize;     
                } __attribute__((packed));

//do a little trick for getting long name characters from a DirEntry
//DirEntryBuffer later gets the same adress as DirEntry
struct DirEntryBuffer {
                 unsigned char longchars[sizeof(struct DirEntry)];
                } __attribute__((packed));

//Prototypes
extern unsigned char GetDriveInformation(void);
extern void UpdateFATBuffer(unsigned long newsector);

#ifdef USE_FAT32
 extern unsigned long GetFirstSectorOfCluster(unsigned long n);
 extern unsigned long GetNextClusterNumber(unsigned long cluster);
 extern unsigned char WriteClusterNumber(unsigned long cluster, unsigned long number);
 extern unsigned long AllocCluster(unsigned long currentcluster);
 extern unsigned long FindFreeCluster(unsigned long currentcluster);

 extern unsigned long endofclusterchain;
 extern unsigned long maxcluster;        // last usable cluster+1
 extern unsigned long FAT32RootCluster;
#else
 extern unsigned long GetFirstSectorOfCluster(U16 n);
 extern U16 GetNextClusterNumber(U16 cluster);
 extern unsigned char WriteClusterNumber(U16 cluster, U16 number);
 extern U16 AllocCluster(U16 currentcluster);
 extern U16 FindFreeCluster(U16 currentcluster);

 extern U16 endofclusterchain;
 extern U16 maxcluster;        // last usable cluster+1
#endif

extern unsigned long maxsect;           // last sector on drive
extern unsigned char secPerCluster;
extern unsigned long BytesPerCluster;

extern unsigned char fatbuf[];   //buffer for FAT sectors
extern unsigned char iob[];      //file i/o buffer

extern unsigned long FATFirstSector;
extern unsigned long FATCurrentSector;
extern unsigned char FATtype;
extern unsigned char FATStatus; // only for FAT write buffering
extern unsigned char iob_status;// only without FAT write buffering 

#ifdef USE_FAT32
  #ifdef FAT_DEBUG_RW_HITS
   extern unsigned long FATWRHits;
   extern unsigned long FATRDHits;
  #endif

 extern unsigned long FirstDirCluster;
 extern unsigned long FileFirstCluster;
 extern unsigned long FileCurrentCluster;
#else
  #ifdef FAT_DEBUG_RW_HITS
   extern U16 FATWRHits;
   extern U16 FATRDHits;
  #endif

 extern U16 FirstDirCluster;
 extern U16 FileFirstCluster;
 extern U16 FileCurrentCluster;
#endif

extern unsigned long FileCurrentSector;
extern unsigned long File1stClusterSector;
extern unsigned long FileClusterCount;
extern unsigned long FileDirSector;
extern unsigned char FileDirOffset;
extern unsigned long FileSize;
extern unsigned long FilePosition;
extern unsigned char FileFlag;
extern unsigned char FileAttr;
extern unsigned char FileName[];
extern unsigned char FileExt[];

extern unsigned long FirstRootSector;
extern unsigned long FirstDataSector; 
extern unsigned long RootDirSectors;

#ifdef USE_FATBUFFER
 #define WriteFATSector(a,b) 	WriteSector((a),(b))
 #define WriteDirSector(a,b) 	WriteSector((a),(b))
 #define WriteFileSector(a,b) 	WriteSector((a),(b))

 #define ReadFATSector(a,b) 	ReadSector((a),(b))
 #define ReadDirSector(a,b) 	ReadSector((a),(b))
 #define ReadFileSector(a,b) 	ReadSector((a),(b))
#else
 extern unsigned char WriteFATSector(unsigned long sector, unsigned char *buf);
 extern unsigned char WriteDirSector(unsigned long sector, unsigned char *buf);
 extern unsigned char WriteFileSector(unsigned long sector, unsigned char *buf);

 extern unsigned char ReadFATSector(unsigned long sector, unsigned char *buf);
 extern unsigned char ReadDirSector(unsigned long sector, unsigned char *buf);
 extern unsigned char ReadFileSector(unsigned long sector, unsigned char *buf);
#endif

#endif //FAT_H
